"""
The frame module is the foundation of the user interface.  The parent class is
the BaseFrame object, which defines the user interface.  User interfaces
(e.g., a web interface, a desktop GUI or a console curses interface) implement
this interface.
"""