'''
Created on Apr 24, 2017

@author: neil
'''
# Locations
LOC_JUNEAU = [58.3019, -134.4197]
LOC_MANAUS = [-3.117034, -60.025780]
LOC_UCLA = [34.07, -118.45] 

# Map colorings.
COLOR_COASTLINES = '#666666'
COLOR_CONTINENTS = '#cccccc'