"""
Module: plotter
---------------

A collection of classes for creating and working with plots.  This package
contains classes for handling scatter plots and plots of data on latitude and
longitude coordinates.

The plotting library is built on Matplotlib.pyplot and the mpltoolkit Basemap.
Instead of simply linking to these libraries, this package implements a small
subset of the capability of these libraries and integrates them into a user
interface.

the projections supported are orthographic, mercator, lambert conformal and
lambert azimuthal (equal area).
"""
